/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it
 * is applied to this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * AgroSense. If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.sensor.greenseeker;

import com.linuxense.javadbf.DBFException;
import com.linuxense.javadbf.DBFReader;
import com.vividsolutions.jts.geom.Coordinate;
import eu.agrosense.client.grid.Grid;
import eu.agrosense.client.grid.GridProvider;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.EnumMap;
import java.util.List;
import javax.swing.Icon;
import nl.cloudfarming.client.farm.model.FarmManager;
import nl.cloudfarming.client.imp.ImportProcessor;
import nl.cloudfarming.client.logging.AppLogFactory;
import nl.cloudfarming.client.logging.AppLogger;
import nl.cloudfarming.client.model.EnterpriseIdProvider;
import nl.cloudfarming.client.model.SensorMetadata;
import nl.cloudfarming.client.model.SensorObservation;
import nl.cloudfarming.client.sensor.AbstractSensorImportProcessor;
import nl.cloudfarming.client.sensor.SensorDataObject;
import nl.cloudfarming.client.sensor.SensorEventKey;
import nl.cloudfarming.client.sensor.SensorPalette;
import nl.cloudfarming.eventbus.BaseGuiEventProducer;
import org.netbeans.api.progress.ProgressHandle;
import org.netbeans.api.progress.ProgressHandleFactory;
import org.netbeans.api.project.Project;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileUtil;
import org.openide.util.Exceptions;
import org.openide.util.ImageUtilities;
import org.openide.util.Lookup;
import org.openide.util.NbBundle;
import org.openide.util.lookup.ServiceProvider;

/**
 * Import processor for the Greenseeker files
 *
 * @author Merijn Zengers <Merijn.Zengers@gmail.nl>
 */
@NbBundle.Messages("sensor greenseeker format description=Greenseeker Sensor file (*.dbf)")
@ServiceProvider(service = ImportProcessor.class)
public class GreenseekerImportProcessor extends AbstractSensorImportProcessor {

    // Messages
    private static final String LOGKEY_TYPE_DISPLAYNAME = NbBundle.getMessage(GreenseekerImportProcessor.class, "sensor.greenseeker.fileprocessor.normal.log.displayname");
    private static final String LOGKEY_FAILED_TO_READ_FILE = NbBundle.getMessage(SensorDataObject.class, "sensor.common.fileprocessor.logerror.failed_to_read_file");
    private static final String LOGKEY_PROCESSING_SUMMARY_PROCESSEDMSG = NbBundle.getMessage(SensorDataObject.class, "sensor.common.fileprocessor.loginfo.nr_of_processed");
    private static final String LOGKEY_PROCESSING_SUMMARY_ERRORMSG = NbBundle.getMessage(SensorDataObject.class, "sensor.common.fileprocessor.logerror.nr_of_records_failed");
    private static final String LOGKEY_PROCESSING_START = NbBundle.getMessage(SensorDataObject.class, "sensor.common.fileprocessor.loginfo.process_start");
    // Other variable declarations
    private static final AppLogger LOG = AppLogFactory.getLogger(GreenseekerImportProcessor.class);
    private final ProgressHandle progressHandle;
    private final EventProducer eventProducer = new EventProducer();
    public static final String GREENSEEKER_FOLDER = "Greenseeker";
    private static final Icon ICON = ImageUtilities.loadImageIcon(GreenseekerDataObject.ICON_BASE, true);
    public static final int NO_FIELDS_BORNE = 20;
    public static final int NO_FIELDS_NORMAL = 14;

    public GreenseekerImportProcessor() {
        this.progressHandle = ProgressHandleFactory.createHandle(NbBundle.getMessage(SensorDataObject.class, "sensor.common.fileprocessor.loginfo.process_description", getFormatDescription()));
    }

    @Override
    public String[] getSupportedExtentions() {
        return new String[]{"dbf"};
    }


    @Override
    public boolean isSupported(FileObject importFile) {
        return GreenseekerDataObject.MIME_TYPE.equals(importFile.getMIMEType());
    }

    @Override
    public String getFormatDescription() {
        return Bundle.sensor_greenseeker_format_description();
    }

    @Override
    public Icon getIcon() {
        return ICON;
    }

    @Override
    public void process(FileObject importFile, Project project) throws IllegalArgumentException {
        process(importFile, project, new ArrayList());
    }

    @Override
    public void process(FileObject importFile, Project project, List toPopulate) throws IllegalArgumentException {
        
        Parser parser;
        switch(fieldCount(importFile)){
            case NO_FIELDS_BORNE:parser = new BorneParser();break;
            default: parser = new NormalParser();break;
        }
        
        List<SensorObservation> sensorData = new ArrayList<>();
        try {
            int success = 0;
            int fail = 0;
            int progress = 0;

            LOG.info(LOGKEY_PROCESSING_START, getLogDisplayName(), importFile.getName());

            try (InputStream inputStream = new FileInputStream(FileUtil.toFile(importFile))) {
                DBFReader reader = new DBFReader(inputStream);

                Object[] dbfSensorRow;
                // Start progressbar
                this.progressHandle.start(reader.getRecordCount());


                while ((dbfSensorRow = reader.nextRecord()) != null) {

                    try {
                        sensorData.add(parser.parseLine(dbfSensorRow));
                        success++;
                    } catch (GreenseekerParseException ex) {
                        LOG.trace(ex);
                        fail++;
                    }
                    progressHandle.progress(++progress);
                }
                LOG.info(LOGKEY_PROCESSING_SUMMARY_PROCESSEDMSG, success);
                if (fail > 0) {
                    LOG.warning(LOGKEY_PROCESSING_SUMMARY_ERRORMSG, fail);
                }
                eventProducer.triggerEvent(SensorEventKey.SENSOR_DATA_IMPORTED);
                progressHandle.finish();
            }
        } catch (Exception ex) {
            LOG.error(LOGKEY_FAILED_TO_READ_FILE, importFile.getName());
            LOG.trace(ex);
        }
        assert (!sensorData.isEmpty() && sensorData.size() >= 1);

        double minx = sensorData.get(0).getLongitude();
        double maxx = sensorData.get(0).getLongitude();
        double miny = sensorData.get(0).getLatitude();
        double maxy = sensorData.get(0).getLatitude();

        for (SensorObservation sd : sensorData) {
            minx = (minx < sd.getLongitude()) ? minx : sd.getLongitude();
            miny = (miny < sd.getLatitude()) ? miny : sd.getLatitude();
            maxx = (maxx > sd.getLongitude()) ? maxx : sd.getLongitude();
            maxy = (maxy > sd.getLatitude()) ? maxy : sd.getLatitude();
        }

        GridProvider gridProvider = Lookup.getDefault().lookup(GridProvider.class);
        
        Grid grid = gridProvider.getDefaultGridSection(project.getLookup().lookup(EnterpriseIdProvider.class),new Coordinate(minx, miny), new Coordinate(maxx, maxy));

        for (SensorObservation sd : sensorData) {
            //TODO sample data gets casted into an int, so multiply with 1000 for example
            if (sd.getNdvi() != null) {
                grid.setRasterValue(new Coordinate(sd.getLongitude(), sd.getLatitude()), (int) (sd.getNdvi() * SensorPalette.SENSOR_DATA_MULTIPLIER));
            }

            if (sd.getWdvi() != null) {
                grid.setRasterValue(new Coordinate(sd.getLongitude(), sd.getLatitude()), (int) (sd.getWdvi() * SensorPalette.SENSOR_DATA_MULTIPLIER));
            }
        }
        toPopulate.add(grid);
    }

    private int fieldCount(FileObject fo) {
        int count = 0;
        if (FileUtil.toFile(fo) != null) { // Template file does not actually exist yet.
            try (InputStream inputStream = fo.getInputStream()) {
                DBFReader reader = new DBFReader(inputStream);
                count = reader.getFieldCount();
            } catch (DBFException ex) {
                LOG.warning(NbBundle.getMessage(GreenseekerDataObject.class, "GreenseekerDataObject.logerror.failed_to_read_dbf", fo.getName()));
                Exceptions.printStackTrace(ex);
            } catch (IOException ex) {
                LOG.warning(NbBundle.getMessage(GreenseekerDataObject.class, "GreenseekerDataObject.logerror.failed_to_read_inputstream", fo.getName()));
                Exceptions.printStackTrace(ex);
            }
        }
        return count;
    }
    
    protected String getLogDisplayName() {
        return LOGKEY_TYPE_DISPLAYNAME;
    }    

    @Override
    protected FileObject getSensorImportSubFolder(Project project) throws IOException {
        return getSubFolder(getSensorImportRootFolder(project), GREENSEEKER_FOLDER);
    }

    /**
     * Generic greenseeker method to build a Greenseeker specific metadata
     * object
     *
     * @return Greenseeker metadata object
     */
    private SensorMetadata buildMetadataFromPreferences() {
        String fileFormat = GreenseekerPreference.SENSOR_DATA_FILE_FORMAT.getValue();
        int sensorHeigth = Integer.parseInt(GreenseekerPreference.GREENSEEKER_SENSOR_HEIGTH.getValue());
        int numberOfSensors = Integer.parseInt(GreenseekerPreference.NUMBER_OF_GREENSEEKER_SENSORS.getValue());
        String greenseekerSensorSpread = GreenseekerPreference.GREENSEEKER_SENSOR_SPREAD.getValue();
        String sensorType = GreenseekerPreference.SENSOR_TYPE.getValue();
        String spectralChannels = GreenseekerPreference.SPECTRAL_CHANNELS.getValue();
        return new SensorMetadata(fileFormat, sensorHeigth, numberOfSensors, greenseekerSensorSpread, sensorType, spectralChannels);
    }

    private String getFileDateFormat() {
        return GreenseekerPreference.SENSOR_DATA_FILE_FORMAT.getValue();
    }

    private interface Parser {

        GreenseekerSensorData parseLine(Object[] payload) throws GreenseekerParseException;
    }

    private class NormalParser implements Parser {

        /**
         * Parses one record of a normal Greenseeker sensorfile.
         *
         * @param payload The record to be processed.
         * @param dataFile The SensorDataFile belonging to this sensor file.
         * @return A record represented as a Greenseeker sensordata object
         * @throws ParseException Exception occurs when date/time cannot be
         * parsed.
         */
        @Override
        public GreenseekerSensorData parseLine(Object[] payload) throws GreenseekerParseException {

            // First create a entity representation of a file
            //converting the differrent data
            EnumMap<GreenseekerFileDescriptionNormal, Object> map = new EnumMap<>(GreenseekerFileDescriptionNormal.class);
            try {
                for (GreenseekerFileDescriptionNormal greenseekerNormalField : GreenseekerFileDescriptionNormal.values()) {
                    if (greenseekerNormalField.isMandatory()) {
                        map.put(greenseekerNormalField, payload[greenseekerNormalField.getPosition()]);
                    }
                }
            } catch (Exception e) {
                throw new GreenseekerParseException(e);
            }

            GreenseekerSensorData sensorData = new GreenseekerSensorData();

            sensorData.setObjectId((Double) map.get(GreenseekerFileDescriptionNormal.OBJECTID));
            sensorData.setLongitude((Double) map.get(GreenseekerFileDescriptionNormal.LONGITUDE));
            sensorData.setLatitude((Double) map.get(GreenseekerFileDescriptionNormal.LATITUDE));
            sensorData.setSpeed((Double) map.get(GreenseekerFileDescriptionNormal.SPEED));
            sensorData.setHeading((Double) map.get(GreenseekerFileDescriptionNormal.HEADING));
            sensorData.setElevation((Double) map.get(GreenseekerFileDescriptionNormal.ELEVATION));
            sensorData.setNdvi((Double) map.get(GreenseekerFileDescriptionNormal.NDVI));
            sensorData.setVi2nd((Double) map.get(GreenseekerFileDescriptionNormal.VI_2ND));

            sensorData.setSensor10((Double) map.get(GreenseekerFileDescriptionNormal.SENSOR_10));
            sensorData.setSensor19((Double) map.get(GreenseekerFileDescriptionNormal.SENSOR_19));
            sensorData.setSensor23((Double) map.get(GreenseekerFileDescriptionNormal.SENSOR_23));
            sensorData.setSensor24((Double) map.get(GreenseekerFileDescriptionNormal.SENSOR_24));
            sensorData.setSensor26((Double) map.get(GreenseekerFileDescriptionNormal.SENSOR_26));
            sensorData.setSensor73((Double) map.get(GreenseekerFileDescriptionNormal.SENSOR_73));

            return sensorData;
        }
    }

    private class BorneParser implements Parser {

        /**
         * Parses one record of a v/d Borne Greenseeker sensorfile.
         *
         * @param payload The record to be processed.
         * @param dataFile The SensorDataFile belonging to this sensor file.
         * @return A record represented as a Greenseeker sensordata object
         * @throws ParseException Exception occurs when date/time cannot be
         * parsed.
         */
        @Override
        public GreenseekerSensorData parseLine(Object[] payload) throws GreenseekerParseException {

            EnumMap<GreenseekerFileDescriptionBorne, Object> map = new EnumMap<>(GreenseekerFileDescriptionBorne.class);
            try {
                for (GreenseekerFileDescriptionBorne greenseekerBorneField : GreenseekerFileDescriptionBorne.values()) {
                    if (greenseekerBorneField.isMandatory()) {
                        map.put(greenseekerBorneField, payload[greenseekerBorneField.getPosition()]);
                    }
                }
            } catch (Exception e) {
                throw new GreenseekerParseException(e);
            }

            GreenseekerSensorData sensorData = new GreenseekerSensorData();
            sensorData.setObjectId((Double) map.get(GreenseekerFileDescriptionBorne.OBJECTID));
            sensorData.setLongitude((Double) map.get(GreenseekerFileDescriptionBorne.LONGITUDE));
            sensorData.setLatitude((Double) map.get(GreenseekerFileDescriptionBorne.LATITUDE));
            sensorData.setSpeed((Double) map.get(GreenseekerFileDescriptionBorne.SPEED));
            sensorData.setHeading((Double) map.get(GreenseekerFileDescriptionBorne.HEADING));
            sensorData.setElevation((Double) map.get(GreenseekerFileDescriptionBorne.ELEVATION));

            sensorData.setNdvi((Double) map.get(GreenseekerFileDescriptionBorne.NDVI));
            sensorData.setVi2nd((Double) map.get(GreenseekerFileDescriptionBorne.VI2ND));

            sensorData.setRed54((Double) map.get(GreenseekerFileDescriptionBorne.RED54));
            sensorData.setNir54((Double) map.get(GreenseekerFileDescriptionBorne.NIR54));

            sensorData.setRed55((Double) map.get(GreenseekerFileDescriptionBorne.RED55));
            sensorData.setNir55((Double) map.get(GreenseekerFileDescriptionBorne.NIR55));

            sensorData.setRed57((Double) map.get(GreenseekerFileDescriptionBorne.RED57));
            sensorData.setNir57((Double) map.get(GreenseekerFileDescriptionBorne.NIR57));

            sensorData.setRed53((Double) map.get(GreenseekerFileDescriptionBorne.RED53));
            sensorData.setNir53((Double) map.get(GreenseekerFileDescriptionBorne.NIR53));

            sensorData.setRed56((Double) map.get(GreenseekerFileDescriptionBorne.RED56));
            sensorData.setNir56((Double) map.get(GreenseekerFileDescriptionBorne.NIR56));
            sensorData.setRed58((Double) map.get(GreenseekerFileDescriptionBorne.RED58));
            sensorData.setNir58((Double) map.get(GreenseekerFileDescriptionBorne.NIR58));

            return sensorData;
        }
    }

    private class EventProducer extends BaseGuiEventProducer {

        public EventProducer() {
            super(SensorEventKey.class);
        }
    }
}
